@@ -27,19 +27,25 @@ __copyright__ = "Copyright (c) 2005-2008 Sun Microsystems Inc." \
                 "Copyright (c) 2011-2016 Igalia, S.L."
 __license__   = "LGPL"
 
+import gi
+gi.require_version("Atspi", "2.0")
+from gi.repository import Atspi
+
 import math
-import pyatspi
 import time
 from gi.repository import Gdk
 from gi.repository import GLib
 
 from . import debug
+from . import focus_manager
 from . import keybindings
 from . import keynames
 from . import messages
 from . import orca_state
 from . import script_manager
 from . import settings
+from .ax_object import AXObject
+from .ax_utilities import AXUtilities
 
 KEYBOARD_EVENT     = "keyboard"
 BRAILLE_EVENT      = "braille"
@@ -64,8 +70,27 @@ class InputEvent:
 
         pass
 
+    def asSingleLineString(self):
+        """Returns a single-line string representation of this event."""
+
+        return f"{self.type}"
+
+def _getXkbStickyKeysState():
+    from subprocess import check_output
+
+    try:
+        output = check_output(['xkbset', 'q'])
+        for line in output.decode('ASCII', errors='ignore').split('\n'):
+            if line.startswith('Sticky-Keys = '):
+                return line.endswith('On')
+    except Exception:
+        pass
+    return False
+
 class KeyboardEvent(InputEvent):
 
+    stickyKeys = _getXkbStickyKeysState()
+
     duplicateCount = 0
     orcaModifierPressed = False
 
@@ -77,6 +102,8 @@ class KeyboardEvent(InputEvent):
     currentOrcaModifierAloneTime = None
     # When the second orca press happened
     secondOrcaModifierTime = None
+    # Sticky modifiers state, to be applied to the next keyboard event
+    orcaStickyModifiers = 0
 
     TYPE_UNKNOWN          = "unknown"
     TYPE_PRINTABLE        = "printable"
@@ -204,27 +231,39 @@ class KeyboardEvent(InputEvent):
                                 Gdk.KEY_Yacute,
                                 Gdk.KEY_yacute]
 
-    def __init__(self, event):
+    def __init__(self, pressed, keycode, keysym, modifiers, text):
         """Creates a new InputEvent of type KEYBOARD_EVENT.
 
         Arguments:
-        - event: the AT-SPI keyboard event
+        - pressed: True if this is a key press, False for a release.
+        - keycode: the hardware keycode.
+        - keysym: the translated keysym.
+        - modifiers: a bitflag giving the active modifiers.
+        - text: the text that would be inserted if this key is pressed.
         """
 
         super().__init__(KEYBOARD_EVENT)
-        self.id = event.id
-        self.type = event.type
-        self.hw_code = event.hw_code
-        self.modifiers = event.modifiers & Gdk.ModifierType.MODIFIER_MASK
-        self.event_string = event.event_string
-        self.keyval_name = Gdk.keyval_name(event.id)
-        self.timestamp = event.timestamp
+        self.id = keysym
+        if pressed:
+            self.type = Atspi.EventType.KEY_PRESSED_EVENT
+        else:
+            self.type = Atspi.EventType.KEY_RELEASED_EVENT
+        self.hw_code = keycode
+        self.modifiers = modifiers & Gdk.ModifierType.MODIFIER_MASK
+        if modifiers & (1 << Atspi.ModifierType.NUMLOCK):
+            self.modifiers |= (1 << Atspi.ModifierType.NUMLOCK)
+        self.event_string = text
+        self.keyval_name = Gdk.keyval_name(keysym)
+        if self.event_string  == "" or self.event_string == " ":
+            self.event_string = self.keyval_name
+        self.timestamp = time.time()
         self.is_duplicate = self in [orca_state.lastInputEvent,
                                      orca_state.lastNonModifierKeyEvent]
-        self._script = orca_state.activeScript
+        self._script = None
         self._app = None
-        self._window = orca_state.activeWindow
-        self._obj = orca_state.locusOfFocus
+        self._window = None
+        self._obj = None
+        self._obj_after_consuming = None
         self._handler = None
         self._consumer = None
         self._should_consume = None
@@ -232,6 +271,7 @@ class KeyboardEvent(InputEvent):
         self._did_consume = None
         self._result_reason = None
         self._bypassOrca = None
+        self._is_kp_with_numlock = False
 
         # Some implementors don't populate this field at all. More often than not,
         # the event_string and the keyval_name coincide for input events.
@@ -245,18 +285,47 @@ class KeyboardEvent(InputEvent):
                 self.id in KeyboardEvent.GDK_ACCENTED_LETTER_KEYS):
             self.event_string = chr(self.id)
 
-        if self._script:
-            self._app = self._script.app
-            if not self._window:
-                self._window = orca_state.activeWindow = self._script.utilities.activeWindow()
-                msg = 'INPUT EVENT: Updated window and active window to %s' % self._window
-                debug.println(debug.LEVEL_INFO, msg, True)
+        # Some implementors don't include numlock in the modifiers. Unfortunately,
+        # trying to heuristically hack around this just by looking at the event
+        # is not reliable. Ditto regarding asking Gdk for the numlock state.
+        if self.keyval_name.startswith("KP"):
+            if self.modifiers & (1 << Atspi.ModifierType.NUMLOCK):
+                self._is_kp_with_numlock = True
 
-        if self._window and self._app != self._window.getApplication():
-            self._script = script_manager.getManager().getScript(self._window.getApplication())
-            self._app = self._script.app
-            msg = 'INPUT EVENT: Updated script to %s' % self._script
-            debug.println(debug.LEVEL_INFO, msg, True)
+        # We typically do little to nothing in the case of a key release. Therefore skip doing
+        # this work.
+        if pressed:
+            self._script = script_manager.getManager().getActiveScript()
+            self._window = focus_manager.getManager().get_active_window()
+            if self._script:
+                self._app = self._script.app
+                if not focus_manager.getManager().can_be_active_window(self._window):
+                    self._window = focus_manager.getManager().find_active_window()
+                    tokens = ["INPUT EVENT: Updating window and active window to", self._window]
+                    debug.printTokens(debug.LEVEL_INFO, tokens, True)
+                    focus_manager.getManager().set_active_window(self._window)
+
+            # We set this after getting the window because changing the window can cause focus to
+            # be updated if the current locus of focus is not in the window we just set as active.
+            self._obj = focus_manager.getManager().get_locus_of_focus()
+
+            if self._window and self._app != AXObject.get_application(self._window):
+                self._script = script_manager.getManager().getScript(
+                    AXObject.get_application(self._window))
+                self._app = self._script.app
+                tokens = ["INPUT EVENT: Updated script to", self._script]
+                debug.printTokens(debug.LEVEL_INFO, tokens, True)
+
+        elif self._isReleaseForLastNonModifierKeyEvent():
+            self._script = orca_state.lastNonModifierKeyEvent._script
+            self._window = orca_state.lastNonModifierKeyEvent._window
+            self._obj = orca_state.lastNonModifierKeyEvent._obj
+            self._obj_after_consuming = orca_state.lastNonModifierKeyEvent._obj_after_consuming
+        else:
+            self._script = script_manager.getManager().getActiveScript()
+            self._window = focus_manager.getManager().get_active_window()
+            self._obj = focus_manager.getManager().get_locus_of_focus()
+            self._obj_after_consuming = self._obj
 
         if self.is_duplicate:
             KeyboardEvent.duplicateCount += 1
@@ -265,13 +334,17 @@ class KeyboardEvent(InputEvent):
 
         self.keyType = None
 
-        _isPressed = event.type == pyatspi.KEY_PRESSED_EVENT
+        role = AXObject.get_role(self._obj)
+        _mayEcho = pressed or role == Atspi.Role.TERMINAL
 
-        try:
-            role = self._obj.getRole()
-        except:
-            role = None
-        _mayEcho = _isPressed or role == pyatspi.ROLE_TERMINAL
+        if KeyboardEvent.stickyKeys and not self.isOrcaModifier() \
+           and not KeyboardEvent.lastOrcaModifierAlone:
+            doubleEvent = self._getDoubleClickCandidate()
+            if doubleEvent and \
+               doubleEvent.modifiers & keybindings.ORCA_MODIFIER_MASK:
+                # this is the second event of a double-click, and sticky Orca
+                # affected the first, so copy over the modifiers to the second
+                KeyboardEvent.orcaStickyModifiers = doubleEvent.modifiers
 
         if not self.isOrcaModifier():
             if KeyboardEvent.orcaModifierPressed:
@@ -290,25 +363,28 @@ class KeyboardEvent(InputEvent):
         elif self.isModifierKey():
             self.keyType = KeyboardEvent.TYPE_MODIFIER
             self.shouldEcho = _mayEcho and settings.enableModifierKeys
-            if self.isOrcaModifier():
+            if self.isOrcaModifier() and not self.is_duplicate:
                 now = time.time()
                 if KeyboardEvent.lastOrcaModifierAlone:
-                    if _isPressed:
+                    if pressed:
                         KeyboardEvent.secondOrcaModifierTime = now
-                    if KeyboardEvent.secondOrcaModifierTime < KeyboardEvent.lastOrcaModifierAloneTime + 0.5:
+                    if (KeyboardEvent.secondOrcaModifierTime <
+                        KeyboardEvent.lastOrcaModifierAloneTime + 0.5):
                         # double-orca, let the real action happen
                         self._bypassOrca = True
-                    if not _isPressed:
+                    if not pressed:
                         KeyboardEvent.lastOrcaModifierAlone = False
                         KeyboardEvent.lastOrcaModifierAloneTime = False
                 else:
-                    KeyboardEvent.orcaModifierPressed = _isPressed
-                    if _isPressed:
+                    KeyboardEvent.orcaModifierPressed = pressed
+                    if pressed:
                         KeyboardEvent.currentOrcaModifierAlone = True
                         KeyboardEvent.currentOrcaModifierAloneTime = now
                     else:
-                        KeyboardEvent.lastOrcaModifierAlone = KeyboardEvent.currentOrcaModifierAlone
-                        KeyboardEvent.lastOrcaModifierAloneTime = KeyboardEvent.currentOrcaModifierAloneTime
+                        KeyboardEvent.lastOrcaModifierAlone = \
+                            KeyboardEvent.currentOrcaModifierAlone
+                        KeyboardEvent.lastOrcaModifierAloneTime = \
+                            KeyboardEvent.currentOrcaModifierAloneTime
         elif self.isFunctionKey():
             self.keyType = KeyboardEvent.TYPE_FUNCTION
             self.shouldEcho = _mayEcho and settings.enableFunctionKeys
@@ -320,7 +396,7 @@ class KeyboardEvent(InputEvent):
             self.shouldEcho = settings.presentLockingKeys
             if self.shouldEcho is None:
                 self.shouldEcho = not settings.onlySpeakDisplayedText
-            self.shouldEcho = self.shouldEcho and _isPressed
+            self.shouldEcho = self.shouldEcho and pressed
         elif self.isAlphabeticKey():
             self.keyType = KeyboardEvent.TYPE_ALPHABETIC
             self.shouldEcho = _mayEcho \
@@ -347,32 +423,57 @@ class KeyboardEvent(InputEvent):
         if not self.isModifierKey():
             self.setClickCount()
 
-        if orca_state.bypassNextCommand and _isPressed:
+        if orca_state.bypassNextCommand and pressed:
             KeyboardEvent.orcaModifierPressed = False
 
         if KeyboardEvent.orcaModifierPressed:
             self.modifiers |= keybindings.ORCA_MODIFIER_MASK
 
+        if KeyboardEvent.stickyKeys:
+            # apply all recorded sticky modifiers
+            self.modifiers |= KeyboardEvent.orcaStickyModifiers
+            if self.isModifierKey():
+                # add this modifier to the sticky ones
+                KeyboardEvent.orcaStickyModifiers |= self.modifiers
+            else:
+                # Non-modifier key, so clear the sticky modifiers. If the user
+                # actually double-presses that key, the modifiers of this event
+                # will be copied over to the second event, see earlier in this
+                # function.
+                KeyboardEvent.orcaStickyModifiers = 0
+
         self._should_consume, self._consume_reason = self.shouldConsume()
 
+    def _getDoubleClickCandidate(self):
+        lastEvent = orca_state.lastNonModifierKeyEvent
+        if isinstance(lastEvent, KeyboardEvent) \
+           and lastEvent.event_string == self.event_string \
+           and self.time - lastEvent.time <= settings.doubleClickTimeout:
+            return lastEvent
+        return None
+
     def setClickCount(self):
         """Updates the count of the number of clicks a user has made."""
 
-        lastEvent = orca_state.lastNonModifierKeyEvent
-        if not isinstance(lastEvent, KeyboardEvent) \
-           or lastEvent.event_string != self.event_string \
-           or self.time - lastEvent.time > settings.doubleClickTimeout:
+        doubleEvent = self._getDoubleClickCandidate()
+        if not doubleEvent:
             self._clickCount = 1
             return
 
-        self._clickCount = lastEvent.getClickCount()
+        self._clickCount = doubleEvent.getClickCount()
         if self.is_duplicate:
             return
 
-        if self.type == pyatspi.KEY_RELEASED_EVENT:
+        if self.type == Atspi.EventType.KEY_RELEASED_EVENT:
             return
 
         if self._clickCount < 3:
+            if doubleEvent._obj != doubleEvent._obj_after_consuming:
+                tokens = ["KEYBOARD EVENT: Resetting click count due to focus change from",
+                          doubleEvent._obj, "to", doubleEvent._obj_after_consuming]
+                debug.printTokens(debug.LEVEL_INFO, tokens, True)
+                self._clickCount = 1
+                return
             self._clickCount += 1
             return
 
@@ -398,20 +499,31 @@ class KeyboardEvent(InputEvent):
             keyval_name = self.keyval_name
             key_type = self.keyType
 
-        return ("KEYBOARD_EVENT:  type=%s\n" % self.type.value_name.upper()) \
-             + ("                 id=%s\n" % keyid) \
-             + ("                 hw_code=%s\n" % hw_code) \
-             + ("                 modifiers=%s\n" % modifiers) \
-             + ("                 event_string=(%s)\n" % event_string) \
-             + ("                 keyval_name=(%s)\n" % keyval_name) \
+        return (f"KEYBOARD_EVENT:  type={self.type.value_name.upper()}\n") \
+             + f"                 id={keyid}\n" \
+             + f"                 hw_code={hw_code}\n" \
+             + f"                 modifiers={modifiers}\n" \
+             + f"                 event_string=({event_string})\n" \
+             + f"                 keyval_name=({keyval_name})\n" \
              + ("                 timestamp=%d\n" % self.timestamp) \
-             + ("                 time=%f\n" % time.time()) \
-             + ("                 keyType=%s\n" % key_type) \
-             + ("                 clickCount=%s\n" % self._clickCount) \
-             + ("                 shouldEcho=%s\n" % self.shouldEcho)
+             + f"                 time={time.time():f}\n" \
+             + f"                 keyType={key_type}\n" \
+             + f"                 clickCount={self._clickCount}\n" \
+             + f"                 shouldEcho={self.shouldEcho}\n"
+
+    def asSingleLineString(self):
+        """Returns a single-line string representation of this event."""
+
+        if self._shouldObscure():
+            return "(obscured)"
+
+        return (
+            f"'{self.event_string}' ({self.keyval_name}) mods: {self.modifiers} "
+            f"{self.type.value_nick}"
+        )
 
     def _shouldObscure(self):
-        if not (self._obj and self._obj.getRole() == pyatspi.ROLE_PASSWORD_TEXT):
+        if not AXUtilities.is_password_text(self._obj):
             return False
 
         if not self.isPrintableKey():
@@ -509,11 +621,11 @@ class KeyboardEvent(InputEvent):
             return self.keyType in KeyboardEvent.TYPE_LOCKING
 
         lockingKeys = ["Caps_Lock", "Shift_Lock", "Num_Lock", "Scroll_Lock"]
-        if not self.event_string in lockingKeys:
+        if self.event_string not in lockingKeys:
             return False
 
         if not orca_state.bypassNextCommand and not self._bypassOrca:
-            return not self.event_string in settings.orcaModifierKeys
+            return self.event_string not in settings.orcaModifierKeys
 
         return True
 
@@ -566,6 +678,11 @@ class KeyboardEvent(InputEvent):
 
         return self.modifiers & keybindings.ORCA_MODIFIER_MASK
 
+    def isKeyPadKeyWithNumlockOn(self):
+        """Return True if this is a key pad key with numlock on."""
+
+        return self._is_kp_with_numlock
+
     def isPrintableKey(self):
         """Return True if this is a printable key."""
 
@@ -580,7 +697,7 @@ class KeyboardEvent(InputEvent):
     def isPressedKey(self):
         """Returns True if the key is pressed"""
 
-        return self.type == pyatspi.KEY_PRESSED_EVENT
+        return self.type == Atspi.EventType.KEY_PRESSED_EVENT
 
     def isPunctuationKey(self):
         """Return True if this is a punctuation key."""
@@ -617,10 +734,7 @@ class KeyboardEvent(InputEvent):
         if not self.isPrintableKey():
             return False
 
-        if orca_state.learnModeEnabled:
-            return False
-
-        script = orca_state.activeScript
+        script = script_manager.getManager().getActiveScript()
         return script and script.utilities.willEchoCharacter(self)
 
     def getLockingState(self):
@@ -632,11 +746,11 @@ class KeyboardEvent(InputEvent):
             return None
 
         if self.event_string == "Caps_Lock":
-            mod = pyatspi.MODIFIER_SHIFTLOCK
+            mod = Atspi.ModifierType.SHIFTLOCK
         elif self.event_string == "Shift_Lock":
-            mod = pyatspi.MODIFIER_SHIFT
+            mod = Atspi.ModifierType.SHIFT
         elif self.event_string == "Num_Lock":
-            mod = pyatspi.MODIFIER_NUMLOCK
+            mod = Atspi.ModifierType.NUMLOCK
         else:
             return None
 
@@ -665,21 +779,26 @@ class KeyboardEvent(InputEvent):
 
         return self._obj
 
+    def getHandler(self):
+        """Returns the handler associated with this key event."""
+
+        return self._handler
+
     def _getUserHandler(self):
         # TODO - JD: This should go away once plugin support is in place.
         try:
             bindings = settings.keyBindingsMap.get(self._script.__module__)
-        except:
+        except Exception:
             bindings = None
         if not bindings:
             try:
                 bindings = settings.keyBindingsMap.get("default")
-            except:
+            except Exception:
                 bindings = None
 
         try:
             handler = bindings.getInputHandler(self)
-        except:
+        except Exception:
             handler = None
 
         return handler
@@ -705,28 +824,13 @@ class KeyboardEvent(InputEvent):
         self._handler = self._getUserHandler() \
             or self._script.keyBindings.getInputHandler(self)
 
-        # TODO - JD: Right now we need to always call consumesKeyboardEvent()
-        # because that method is updating state, even in instances where there
-        # is no handler.
-        scriptConsumes = self._script.consumesKeyboardEvent(self)
+        scriptConsumes = self._handler is not None and self._handler.is_enabled()
 
         if self._isReleaseForLastNonModifierKeyEvent():
             return scriptConsumes, 'Is release for last non-modifier keyevent'
 
-        if orca_state.learnModeEnabled:
-            if self.event_string == 'Escape':
-                self._consumer = self._script.exitLearnMode
-                return True, 'Exiting Learn Mode'
-
-            if self.event_string == 'F1' and not self.modifiers:
-                self._consumer = self._script.showHelp
-                return True, 'Showing Help'
-
-            if self.event_string in ['F2', 'F3'] and not self.modifiers:
-                self._consumer = self._script.listOrcaShortcuts
-                return True, 'Listing shortcuts'
-
-            self._consumer = self._presentHandler
+        if self._script.learnModePresenter.is_active():
+            self._consumer = self._script.learnModePresenter.handle_event
             return True, 'In Learn Mode'
 
         if self.isModifierKey():
@@ -734,23 +838,14 @@ class KeyboardEvent(InputEvent):
                 return False, 'Non-Orca modifier not in Learn Mode'
             return True, 'Orca modifier'
 
-        if orca_state.listNotificationsModeEnabled:
-            self._consumer = self._script.listNotifications
-            return True, 'Listing notifications'
-
         if not self._handler:
             return False, 'No handler'
 
+        if not self._handler.is_enabled():
+            return False, 'Handler is disabled'
+
         return scriptConsumes, 'Script indication'
 
-    def didConsume(self):
-        """Returns True if this event was consumed."""
-
-        if self._did_consume is not None:
-            return self._did_consume
-
-        return False
-
     def isHandledBy(self, method):
         if not self._handler:
             return False
@@ -761,16 +856,10 @@ class KeyboardEvent(InputEvent):
         if self.isPressedKey():
             self._script.presentationInterrupt()
 
-        return self._script.presentKeyboardEvent(self)
-
-    def _presentHandler(self, input_event=None):
-        if not self._handler:
+        if self._script.learnModePresenter.is_active():
             return False
 
-        if self._handler.learnModeEnabled and self._handler.description:
-            self._script.presentMessage(self._handler.description)
-
-        return True
+        return self._script.presentKeyboardEvent(self)
 
     def process(self):
         """Processes this input event."""
@@ -784,32 +873,37 @@ class KeyboardEvent(InputEvent):
         if self.is_duplicate:
             data = '%s DUPLICATE EVENT #%i' % (data, KeyboardEvent.duplicateCount)
 
-        msg = '\nvvvvv PROCESS %s: %s vvvvv' % (self.type.value_name.upper(), data)
-        debug.println(debug.LEVEL_INFO, msg, False)
+        msg = f'\nvvvvv PROCESS {self.type.value_name.upper()}: {data} vvvvv'
+        debug.printMessage(debug.LEVEL_INFO, msg, False)
 
-        msg = 'HOST_APP: %s' % self._app
-        debug.println(debug.LEVEL_INFO, msg, True)
+        tokens = ["HOST_APP:", self._app]
+        debug.printTokens(debug.LEVEL_INFO, tokens, True)
 
-        msg = 'WINDOW:   %s' % self._window
-        debug.println(debug.LEVEL_INFO, msg, True)
+        tokens = ["WINDOW:", self._window]
+        debug.printTokens(debug.LEVEL_INFO, tokens, True)
 
-        msg = 'LOCATION: %s' % self._obj
-        debug.println(debug.LEVEL_INFO, msg, True)
+        tokens = ["LOCATION:", self._obj_after_consuming or self._obj]
+        debug.printTokens(debug.LEVEL_INFO, tokens, True)
 
-        msg = 'CONSUME:  %s (%s)' % (self._should_consume, self._consume_reason)
-        debug.println(debug.LEVEL_INFO, msg, True)
+        tokens = ["CONSUME:", self._should_consume, self._consume_reason]
+        debug.printTokens(debug.LEVEL_INFO, tokens, True)
 
         self._did_consume, self._result_reason = self._process()
+        tokens = ["CONSUMED:", self._did_consume, self._result_reason]
+        debug.printTokens(debug.LEVEL_INFO, tokens, True)
 
-        if self._should_consume != self._did_consume:
-            msg = 'CONSUMED: %s (%s)' % (self._did_consume, self._result_reason)
-            debug.println(debug.LEVEL_INFO, msg, True)
+        script = script_manager.getManager().getActiveScript()
+        if debug.LEVEL_INFO >= debug.debugLevel and script is not None:
+            attributes = script.getTransferableAttributes()
+            for key, value in attributes.items():
+                msg = f"INPUT EVENT: {key}: {value}"
+                debug.printMessage(debug.LEVEL_INFO, msg, True)
 
-        msg = 'TOTAL PROCESSING TIME: %.4f' % (time.time() - startTime)
-        debug.println(debug.LEVEL_INFO, msg, True)
+        msg = f"TOTAL PROCESSING TIME: {time.time() - startTime:.4f}"
+        debug.printMessage(debug.LEVEL_INFO, msg, True)
 
-        msg = '^^^^^ PROCESS %s: %s ^^^^^\n' % (self.type.value_name.upper(), data)
-        debug.println(debug.LEVEL_INFO, msg, False)
+        msg = f"^^^^^ PROCESS {self.type.value_name.upper()}: {data} ^^^^^\n"
+        debug.printMessage(debug.LEVEL_INFO, msg, False)
 
         return self._did_consume
 
@@ -819,7 +913,7 @@ class KeyboardEvent(InputEvent):
         if self._bypassOrca:
             if (self.event_string == "Caps_Lock" \
                 or self.event_string == "Shift_Lock") \
-               and self.type == pyatspi.KEY_PRESSED_EVENT:
+               and self.type == Atspi.EventType.KEY_PRESSED_EVENT:
                     self._lock_mod()
                     self.keyType = KeyboardEvent.TYPE_LOCKING
                     self._present()
@@ -849,6 +943,7 @@ class KeyboardEvent(InputEvent):
         if orca_state.bypassNextCommand:
             if not self.isModifierKey():
                 orca_state.bypassNextCommand = False
+                self._script.addKeyGrabs("bypassed next command")
             return False, 'Bypass next command'
 
         if not self._should_consume:
@@ -868,52 +963,58 @@ class KeyboardEvent(InputEvent):
             def lockit():
                 try:
                     if modifiers & modifier:
-                        lock = pyatspi.KEY_UNLOCKMODIFIERS
-                        debug.println(debug.LEVEL_INFO, "Locking capslock", True)
+                        lock = Atspi.KeySynthType.UNLOCKMODIFIERS
+                        debug.printMessage(debug.LEVEL_INFO, "Unlocking capslock", True)
                     else:
-                        lock = pyatspi.KEY_LOCKMODIFIERS
-                        debug.println(debug.LEVEL_INFO, "Unlocking capslock", True)
-                    pyatspi.Registry.generateKeyboardEvent(modifier, None, lock)
-                    debug.println(debug.LEVEL_INFO, "Done with capslock", True)
-                except:
-                    debug.println(debug.LEVEL_INFO, "Could not trigger capslock, " \
+                        lock = Atspi.KeySynthType.LOCKMODIFIERS
+                        debug.printMessage(debug.LEVEL_INFO, "Locking capslock", True)
+                    Atspi.generate_keyboard_event(modifier, "", lock)
+                    debug.printMessage(debug.LEVEL_INFO, "Done with capslock", True)
+                except Exception:
+                    debug.printMessage(debug.LEVEL_INFO, "Could not trigger capslock, " \
                         "at-spi2-core >= 2.32 is needed for triggering capslock", True)
                     pass
             return lockit
         if self.event_string == "Caps_Lock":
-            modifier = 1 << pyatspi.MODIFIER_SHIFTLOCK
+            modifier = 1 << Atspi.ModifierType.SHIFTLOCK
         elif self.event_string == "Shift_Lock":
-            modifier = 1 << pyatspi.MODIFIER_SHIFT
+            modifier = 1 << Atspi.ModifierType.SHIFT
         else:
-            msg = "Unknown locking key %s" % self.event_string
-            debug.println(debug.LEVEL_WARNING, msg, True)
+            tokens = ["Unknown locking key", self.event_string]
+            debug.printTokens(debug.LEVEL_WARNING, tokens, True)
             return
-        debug.println(debug.LEVEL_INFO, "Scheduling capslock", True)
+        debug.printMessage(debug.LEVEL_INFO, "Scheduling capslock", True)
         GLib.timeout_add(1, lock_mod(self.modifiers, modifier))
 
     def _consume(self):
         startTime = time.time()
         data = "'%s' (%d)" % (self.event_string, self.hw_code)
-        msg = 'vvvvv CONSUME %s: %s vvvvv' % (self.type.value_name.upper(), data)
-        debug.println(debug.LEVEL_INFO, msg, False)
+        msg = f'\nvvvvv CONSUME {self.type.value_name.upper()}: {data} vvvvv'
+        debug.printMessage(debug.LEVEL_INFO, msg, False)
 
         if self._consumer:
-            msg = 'INFO: Consumer is %s' % self._consumer.__name__
-            debug.println(debug.LEVEL_INFO, msg, True)
+            msg = f'KEYBOARD EVENT: Consumer is {self._consumer.__name__}'
+            debug.printMessage(debug.LEVEL_INFO, msg, True)
             self._consumer(self)
         elif self._handler.function:
-            msg = 'INFO: Handler is %s' % self._handler.description
-            debug.println(debug.LEVEL_INFO, msg, True)
+            msg = f'KEYBOARD EVENT: Handler is {self._handler}'
+            debug.printMessage(debug.LEVEL_INFO, msg, True)
             self._handler.function(self._script, self)
         else:
-            msg = 'INFO: No handler or consumer'
-            debug.println(debug.LEVEL_INFO, msg, True)
+            msg = 'KEYBOARD EVENT: No handler or consumer'
+            debug.printMessage(debug.LEVEL_INFO, msg, True)
 
-        msg = 'TOTAL PROCESSING TIME: %.4f' % (time.time() - startTime)
-        debug.println(debug.LEVEL_INFO, msg, True)
+        self._obj_after_consuming = focus_manager.getManager().get_locus_of_focus()
+        if (self._obj != self._obj_after_consuming):
+            tokens = ["KEYBOARD EVENT: Consumer changed focus from", self._obj, "to",
+                      self._obj_after_consuming]
+            debug.printTokens(debug.LEVEL_INFO, tokens, True)
 
-        msg = '^^^^^ CONSUME %s: %s ^^^^^' % (self.type.value_name.upper(), data)
-        debug.println(debug.LEVEL_INFO, msg, False)
+        msg = f'TOTAL PROCESSING TIME: {time.time() - startTime:.4f}'
+        debug.printMessage(debug.LEVEL_INFO, msg, True)
+
+        msg = f'^^^^^ CONSUME {self.type.value_name.upper()}: {data} ^^^^^\n'
+        debug.printMessage(debug.LEVEL_INFO, msg, False)
 
         return False
 
@@ -927,6 +1028,65 @@ class BrailleEvent(InputEvent):
         """
         super().__init__(BRAILLE_EVENT)
         self.event = event
+        self._script = script_manager.getManager().getActiveScript()
+
+    def __str__(self):
+        return f"{self.type.upper()} {self.event}"
+
+    def getHandler(self):
+        """Returns the handler associated with this event."""
+
+        command = self.event["command"]
+        user_bindings = None
+        user_bindings_map = settings.brailleBindingsMap
+        if self._script.name in user_bindings_map:
+            user_bindings = user_bindings_map[self._script.name]
+        elif "default" in user_bindings_map:
+            user_bindings = user_bindings_map["default"]
+
+        if user_bindings and command in user_bindings:
+            handler = user_bindings[command]
+            tokens = [f"BRAILLE EVENT: User handler for command {command} is", handler]
+            debug.printTokens(debug.LEVEL_INFO, tokens, True)
+            return handler
+
+        handler = self._script.brailleBindings.get(command)
+        tokens = [f"BRAILLE EVENT: Handler for command {command} is", handler]
+        debug.printTokens(debug.LEVEL_INFO, tokens, True)
+        return handler
+
+    def process(self):
+        tokens = ["\nvvvvv PROCESS", self, "vvvvv"]
+        debug.printTokens(debug.LEVEL_INFO, tokens, False)
+
+        start_time = time.time()
+        result = self._process()
+        msg = f"TOTAL PROCESSING TIME: {time.time() - start_time:.4f}"
+        debug.printMessage(debug.LEVEL_INFO, msg, False)
+
+        tokens = ["^^^^^ PROCESS", self, "^^^^^"]
+        debug.printTokens(debug.LEVEL_INFO, tokens, False)
+        return result
+
+    def _process(self):
+        handler = self.getHandler()
+        if not handler:
+            if self._script.learnModePresenter.is_active():
+                tokens = ["BRAILLE EVENT: Learn mode presenter handles", self]
+                debug.printTokens(debug.LEVEL_INFO, tokens, True)
+                return True
+
+            tokens = ["BRAILLE EVENT: No handler found for", self]
+            debug.printTokens(debug.LEVEL_INFO, tokens, True)
+            return False
+
+        try:
+            handler.processInputEvent(self._script, self)
+        except Exception as error:
+            tokens = ["BRAILLE EVENT: Exception processing:", self, f": {error}"]
+            debug.printTokens(debug.LEVEL_WARNING, tokens, True)
+
+        return True
 
 class MouseButtonEvent(InputEvent):
 
@@ -934,7 +1094,7 @@ class MouseButtonEvent(InputEvent):
         display = Gdk.Display.get_default()
         seat = Gdk.Display.get_default_seat(display)
         _pointer = seat.get_pointer()
-    except:
+    except Exception:
         _pointer = None
 
     def __init__(self, event):
@@ -945,9 +1105,9 @@ class MouseButtonEvent(InputEvent):
         self.y = event.detail2
         self.pressed = event.type.endswith('p')
         self.button = event.type[len("mouse:button:"):-1]
-        self._script = orca_state.activeScript
-        self.window = orca_state.activeWindow
-        self.obj = None
+        self._script = script_manager.getManager().getActiveScript()
+        self.window = focus_manager.getManager().get_active_window()
+        self.app = None
 
         if self.pressed:
             self._validateCoordinates()
@@ -955,14 +1115,13 @@ class MouseButtonEvent(InputEvent):
         if not self._script:
             return
 
-        if not self._script.utilities.canBeActiveWindow(self.window):
-            self.window = self._script.utilities.activeWindow()
+        if not focus_manager.getManager().can_be_active_window(self.window):
+            self.window = focus_manager.getManager().find_active_window()
 
         if not self.window:
             return
 
-        self.obj = self._script.utilities.descendantAtPoint(
-            self.window, self.x, self.y, event.any_data)
+        self.app = AXObject.get_application(self.window)
 
     def _validateCoordinates(self):
         if not self._pointer:
@@ -972,9 +1131,11 @@ class MouseButtonEvent(InputEvent):
         if math.sqrt((self.x - x)**2 + (self.y - y)**2) < 25:
             return
 
-        msg = "WARNING: Event coordinates (%i, %i) may be bogus. " \
-              "Updating to (%i, %i)." % (self.x, self.y, x, y)
-        debug.println(debug.LEVEL_INFO, msg, True)
+        msg = (
+            f"WARNING: Event coordinates ({self.x}, {self.y}) may be bogus. "
+            f"Updating to ({x}, {y})"
+        )
+        debug.printMessage(debug.LEVEL_INFO, msg, True)
         self.x, self.y = x, y
 
     def setClickCount(self):
@@ -999,7 +1160,7 @@ class MouseButtonEvent(InputEvent):
 
 class InputEventHandler:
 
-    def __init__(self, function, description, learnModeEnabled=True):
+    def __init__(self, function, description, learnModeEnabled=True, enabled=True):
         """Creates a new InputEventHandler instance.  All bindings
         (e.g., key bindings and braille bindings) will be handled
         by an instance of an InputEventHandler.
@@ -1014,11 +1175,14 @@ class InputEventHandler:
         - learnModeEnabled: if True, the description will be spoken and
                             brailled if learn mode is enabled.  If False,
                             the function will be called no matter what.
+        - enabled: Whether this hander can be used, i.e. based on mode, the
+          feature being enabled/active, etc.
         """
 
         self.function = function
         self.description = description
         self.learnModeEnabled = learnModeEnabled
+        self._enabled = enabled
 
     def __eq__(self, other):
         """Compares one input handler to another."""
@@ -1028,12 +1192,23 @@ class InputEventHandler:
 
         return (self.function == other.function)
 
+    def __str__(self):
+        return f"{self.description} (enabled: {self._enabled})"
+
+    def is_enabled(self):
+        """Returns True if this handler is enabled."""
+
+        msg = f"INPUT EVENT HANDLER: {self.description} is enabled: {self._enabled}"
+        debug.printMessage(debug.LEVEL_INFO, msg, True)
+        return self._enabled
+
+    def set_enabled(self, enabled):
+        """Sets this handler's enabled state."""
+
+        self._enabled = enabled
+
     def processInputEvent(self, script, inputEvent):
-        """Processes an input event.  If learnModeEnabled is True,
-        this will merely present the description of the input event via
-        If learnModeEnabled is False, this will call the function bound
-        to this InputEventHandler instance, passing the inputEvent as
-        the sole argument to the function.
+        """Processes an input event.
 
         This function is expected to return True if it consumes the
         event; otherwise it is expected to return False.
@@ -1044,16 +1219,13 @@ class InputEventHandler:
                       to this InputEventHandler instance.
         """
 
-        consumed = False
+        if not self._enabled:
+            return False
 
-        if orca_state.learnModeEnabled and self._learnModeEnabled:
-            if self.description:
-                script.presentMessage(self.description)
-                consumed = True
-        else:
-            try:
-                consumed = self.function(script, inputEvent)
-            except:
-                debug.printException(debug.LEVEL_SEVERE)
+        consumed = False
+        try:
+            consumed = self.function(script, inputEvent)
+        except Exception:
+            debug.printException(debug.LEVEL_SEVERE)
 
         return consumed
