@@ -26,9 +26,11 @@ __date__      = "$Date$"
 __copyright__ = "Copyright (c) 2005-2008 Sun Microsystems Inc."
 __license__   = "LGPL"
 
+import os
+
 # $ORCA_VERSION
 #
-version     = "45.2"
+version     = "46.2"
 
 # The revision if built from git; otherwise an empty string
 #
@@ -44,7 +46,7 @@ package     = "orca"
 
 # The location of the data directory (usually "share").
 #
-datadir = "${prefix}/share".replace('${prefix}', '/usr/local')
+datadir = os.path.join('/usr/local','share')
 
 # The directory where we could find liblouis translation tables.
 #
